import unittest


class SphericalProjectionTestCase(unittest.TestCase):

    @unittest.expectedFailure
    def test_equal_area(self):
        raise NotImplementedError()

    @unittest.expectedFailure
    def test_equal_angle(self):
        raise NotImplementedError()

    @unittest.expectedFailure
    def test__project(self):
        raise NotImplementedError()
